@@ -17,9 +17,11 @@
 # with this program; if not, write to the Free Software Foundation, Inc.,
 # 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
+import datetime
 import getpass
 import os
 import os.path
+import pwd
 import re
 import signal
 import subprocess
@@ -27,7 +29,6 @@ import sys
 import threading
 import time
 import types
-import datetime
 
 try:
     import paramiko as ssh
@@ -36,7 +37,7 @@ except ImportError:
     print("  $ sudo easy_install paramiko")
     sys.exit(1)
 
-__version__ = '0.0.7'
+__version__ = '0.0.8'
 __author__ = 'Christian Vest Hansen'
 __author_email__ = 'karmazilla@gmail.com'
 __url__ = 'http://www.nongnu.org/fab/'
@@ -53,7 +54,7 @@ ENV = {
     'fab_author': __author__,
     'fab_mode': 'rolling',
     'fab_port': 22,
-    'fab_user': os.getlogin(),
+    'fab_user': pwd.getpwuid(os.getuid())[0],
     'fab_password': None,
     'fab_pkey': None,
     'fab_key_filename': None,
@@ -61,6 +62,7 @@ ENV = {
     'fab_shell': '/bin/bash -l -c "%s"',
     'fab_timestamp': datetime.datetime.utcnow().strftime('%F_%H-%M-%S'),
     'fab_print_real_sudo': False,
+    'fab_fail': 'abort',
 }
 
 CONNECTIONS = []
@@ -324,16 +326,23 @@ def run(host, client, env, cmd, **kwargs):
         run("ls")
     
     """
-    cmd = env['fab_shell'] % _lazy_format(cmd, env).replace('"', '\\"')
+    cmd = _lazy_format(cmd, env)
+    real_cmd = env['fab_shell'] % cmd.replace('"', '\\"')
     if not _confirm_proceed('run', host, kwargs):
-        return False # TODO: should we return False in fail??
+        return False
     print("[%s] run: %s" % (host, cmd))
-    stdin, stdout, stderr = client.exec_command(cmd)
+    chan = client._transport.open_session()
+    chan.exec_command(real_cmd)
+    bufsize = -1
+    stdin = chan.makefile('wb', bufsize)
+    stdout = chan.makefile('rb', bufsize)
+    stderr = chan.makefile_stderr('rb', bufsize)
+    
     out_th = _start_outputter("[%s] out" % host, stdout)
     err_th = _start_outputter("[%s] err" % host, stderr)
-    out_th.join()
-    err_th.join()
-    return True
+    status = chan.recv_exit_status()
+    chan.close()
+    return status == 0
 
 @operation
 @run_per_host
@@ -363,16 +372,22 @@ def sudo(host, client, env, cmd, **kwargs):
     if not _confirm_proceed('sudo', host, kwargs):
         return False # TODO: should we return False in fail??
     print("[%s] sudo: %s" % (host, cmd))
-    stdin, stdout, stderr = client.exec_command(real_cmd)
+    chan = client._transport.open_session()
+    chan.exec_command(cmd)
+    bufsize = -1
+    stdin = chan.makefile('wb', bufsize)
+    stdout = chan.makefile('rb', bufsize)
+    stderr = chan.makefile_stderr('rb', bufsize)
     if passwd:
         stdin.write(env['fab_password'])
         stdin.write('\n')
         stdin.flush()
+    
     out_th = _start_outputter("[%s] out" % host, stdout)
     err_th = _start_outputter("[%s] err" % host, stderr)
-    out_th.join()
-    err_th.join()
-    return True
+    status = chan.recv_exit_status()
+    chan.close()
+    return status == 0
 
 @operation
 def local(cmd, **kwargs):
@@ -395,12 +410,7 @@ def local(cmd, **kwargs):
     print("[localhost] run: " + final_cmd)
     retcode = subprocess.call(final_cmd, shell=True)
     if retcode != 0:
-        failcode = _get_failcode(kwargs)
-        if failcode > 1:
-            print("Warning: failed to execute command:")
-            print("\t" + final_cmd)
-        if failcode > 2:
-            sys.exit(1)
+        _fail(kwargs, "Local command failed:\n" + _indent(final_cmd))
 
 @operation
 def local_per_host(cmd, **kwargs):
@@ -411,16 +421,23 @@ def local_per_host(cmd, **kwargs):
     with this operation, the command is executed (and have its variables
     expanded) for each host in fab_hosts.
     
-    This operation is defined to take the same keyword arguments as local().
+    May take an additional 'fail' keyword argument with one of these values:
+        * ignore - do nothing on failure
+        * warn - print warning on failure
+        * abort - terminate fabric on failure
     
     Example:
         local_per_host("scp -i login.key stuff.zip $(fab_host):stuff.zip")
     
     """
     _check_fab_hosts()
-    for host in ENV['fab_hosts']:
-        ENV['fab_host'] = host
-        local(cmd, **kwargs)
+    for host_conn in CONNECTIONS:
+        env = host_conn.get_env()
+        final_cmd = _lazy_format(cmd, env)
+        print(_lazy_format("[localhost/$(fab_host)] run: " + final_cmd, env))
+        retcode = subprocess.call(final_cmd, shell=True)
+        if retcode != 0:
+            _fail(kwargs, "Local command failed:\n" + _indent(final_cmd))
 
 @operation
 def load(filename, **kwargs):
@@ -440,7 +457,6 @@ def load(filename, **kwargs):
         load("conf/production-settings.py")
     
     """
-    failcode = _get_failcode(kwargs)
     if os.path.exists(filename):
         execfile(filename)
         for name, obj in locals().items():
@@ -448,11 +464,9 @@ def load(filename, **kwargs):
                 COMMANDS[name] = obj
             if not name.startswith('_'):
                 __builtins__[name] = obj
-    elif failcode > 1:
-        print("Warning: Cannot load file '%s'." % filename)
-        print("No such file in your current directory.")
-        if failcode > 2:
-            sys.exit(1)
+    else:
+        _fail(kwargs, "Load failed:\n" + _indent(
+            "File not found: " + filename))
 
 @operation
 def upload_project(**kwargs):
@@ -799,7 +813,29 @@ def _set(**kwargs):
 
 @command("shell")
 def _shell(**kwargs):
-    "Start an interactive shell connection to the specified hosts."
+    """
+    Start an interactive shell connection to the specified hosts.
+    
+    Optionally takes a list of hostnames as arguments, if Fabric is, by
+    the time this command runs, not already connected to one or more
+    hosts. If you provide hostnames and Fabric is already connected, then
+    Fabric will, depending on fab_fail, complain and abort.
+    
+    The fab_fail variable can be overwritten with the 'set' command, or
+    by specifying an additional 'fail' argument.
+    
+    Examples:
+        $fab shell
+        $fab shell:localhost,127.0.0.1
+        $fab shell:localhost,127.0.0.1,fail=warn
+    
+    """
+    # expect every arg w/o a value to be a hostname
+    hosts = filter(lambda k: not kwargs[k], kwargs.keys())
+    if hosts:
+        if CONNECTIONS:
+            _fail(kwargs, "Already connected to predefined fab_hosts.")
+        set(fab_hosts = hosts)
     def lines():
         try:
             while True:
@@ -826,34 +862,104 @@ def _fanout_strategy(fn, *args, **kwargs):
     THIS STRATEGY IS CURRENTLY BROKEN!
     
     """
+    err_msg = "The $(fab_current_operation) operation failed on $(fab_host)"
     threads = []
-    success = True
-    for host, client in CONNECTIONS:
-        env = dict(ENV)
-        env['fab_host'] = host
+    for host_conn in CONNECTIONS:
+        env = host_conn.get_env()
+        env['fab_current_operation'] = fn.__name__
+        host = env['fab_host']
+        client = host_conn.client
         def functor():
-            success = success and fn(host, client, env, *args, **kwargs)
+            _try_run_operation(fn, host, client, env, *args, **kwargs)
         thread = threading.Thread(None, functor)
         thread.setDaemon(True)
-        thread.start()
         threads.append(thread)
-    for thread in threads:
-        thread.join()
-    return success
+    map(threading.Thread.start, threads)
+    map(threading.Thread.join, threads)
 
 @strategy("rolling")
 def _rolling_strategy(fn, *args, **kwargs):
     """One-at-a-time fail-fast strategy."""
-    for host, client in CONNECTIONS:
-        env = dict(ENV)
-        env['fab_host'] = host
-        if not fn(host, client, env, *args, **kwargs):
-            return False
-    return True
+    err_msg = "The $(fab_current_operation) operation failed on $(fab_host)"
+    for host_conn in CONNECTIONS:
+        env = host_conn.get_env()
+        env['fab_current_operation'] = fn.__name__
+        host = env['fab_host']
+        client = host_conn.client
+        _try_run_operation(fn, host, client, env, *args, **kwargs)
 
 #
 # Internal plumbing:
 #
+
+class HostConnection(object):
+    """
+    A connection to an SSH host - wraps an SSHClient.
+    
+    Instances of this class populate the CONNECTIONS list.
+    """
+    def __init__(self, hostname, port, global_env, user_local_env):
+        self.global_env = global_env
+        self.user_local_env = user_local_env
+        self.host_local_env = {
+            'fab_host': hostname,
+            'fab_port': port,
+        }
+        self.client = None
+    def get_env(self):
+        "Create a new environment that is the union of local and global envs."
+        env = dict(self.global_env)
+        env.update(self.user_local_env)
+        env.update(self.host_local_env)
+        return env
+    def connect(self):
+        env = self.get_env()
+        new_host_key = env['fab_new_host_key']
+        client = ssh.SSHClient()
+        client.load_system_host_keys()
+        if new_host_key == 'accept':
+            client.set_missing_host_key_policy(ssh.AutoAddPolicy())
+        try:
+            self._do_connect(client, env)
+        except (ssh.AuthenticationException, ssh.SSHException):
+            PASS_PROMPT = \
+                "Password for $(fab_user)@$(fab_host)$(fab_passprompt_suffix)"
+            if 'fab_password' in env and env['fab_password']:
+                env['fab_passprompt_suffix'] = " [Enter for previous]: "
+            else:
+                env['fab_passprompt_suffix'] = ": "
+            connected = False
+            password = None
+            while not connected:
+                try:
+                    password = getpass.getpass(_lazy_format(PASS_PROMPT, env))
+                    env['fab_password'] = password
+                    self._do_connect(client, env)
+                    connected = True
+                except ssh.AuthenticationException:
+                    print("Bad password.")
+                    env['fab_passprompt_suffix'] = ": "
+                except (EOFError, TypeError):
+                    # ctrl-D or ctrl-C on password prompt
+                    print
+                    sys.exit(0)
+            self.host_local_env['fab_password'] = password
+            self.user_local_env['fab_password'] = password
+        self.client = client
+    def disconnect(self):
+        if self.client:
+            self.client.close()
+    def _do_connect(self, client, env):
+        host = env['fab_host']
+        port = env['fab_port']
+        username = env['fab_user']
+        password = env['fab_password']
+        pkey = env['fab_pkey']
+        key_filename = env['fab_key_filename']
+        client.connect(host, port, username, password, pkey, key_filename)
+    def __str__(self):
+        return self.host_local_env['fab_host']
+
 def _indent(text, level=4):
     "Indent all lines in text with 'level' number of spaces, default 4."
     return '\n'.join(((' ' * level) + line for line in text.splitlines()))
@@ -861,7 +967,21 @@ def _indent(text, level=4):
 def _print_help_for(name, doc):
     "Output a pretty-printed help text for the given name & doc"
     default_help_msg = '* No help-text found.'
-    print("Help for '%s':\n%s" % (name, _indent(doc or default_help_msg)))
+    msg = doc or default_help_msg
+    lines = msg.splitlines()
+    # remove leading blank lines
+    while lines and lines[0].strip() == '':
+        lines = lines[1:]
+    # remove trailing blank lines
+    while lines and lines[-1].strip() == '':
+        lines = lines[:-1]
+    if lines:
+        msg = '\n'.join(lines)
+        if not msg.startswith('    '):
+            msg = _indent(msg)
+        print("Help for '%s':\n%s" % (name, msg))
+    else:
+        print("No help message found for '%s'." % name)
 
 def _print_help_for_in(name, dictionary):
     "Print a pretty help text for the named function in the dict."
@@ -888,47 +1008,54 @@ def _check_fab_hosts():
         print("Please set it in your fabfile.")
         print("Example: set(fab_hosts=['node1.com', 'node2.com'])")
         sys.exit(1)
-
-def _connect():
-    "Populate CONNECTIONS with (hostname, client) tuples as per fab_hosts."
-    _check_fab_hosts()
-    signal.signal(signal.SIGINT, lambda: _disconnect() and sys.exit(0))
-    if not get('fab_connected'):
-        print(_lazy_format("Logging into the following hosts as $(fab_user):"))
-        print(_indent('\n'.join(ENV['fab_hosts'])))
-    def_port = ENV['fab_port']
-    username = ENV['fab_user']
-    password = ENV['fab_password']
-    pkey = ENV['fab_pkey']
-    key_filename = ENV['fab_key_filename']
-    for host in ENV['fab_hosts']:
-        host, _, port = partition(host, ':')
-        portnr = int(port or def_port)
-        client = ssh.SSHClient()
-        client.load_system_host_keys()
-        if 'fab_new_host_key' in ENV and ENV['fab_new_host_key'] == 'accept':
-            client.set_missing_host_key_policy(ssh.AutoAddPolicy())
-        try:
-            client.connect(
-                host, portnr, username, password, pkey, key_filename
-            )
-        except (ssh.AuthenticationException, ssh.SSHException):
-            password = ENV['fab_password'] = getpass.getpass()
-            client.connect(
-                host, portnr, username, password, pkey, key_filename
-            )
-        CONNECTIONS.append((host, client))
-    if not CONNECTIONS:
+    if len(ENV['fab_hosts']) == 0:
         print("The fab_hosts list was empty.")
         print("Please specify some hosts to connect to.")
         sys.exit(1)
+
+def _connect():
+    "Populate CONNECTIONS with HostConnection instances as per fab_hosts."
+    _check_fab_hosts()
+    signal.signal(signal.SIGINT, lambda: _disconnect() and sys.exit(0))
+    global CONNECTIONS
+    def_port = ENV['fab_port']
+    username = ENV['fab_user']
+    fab_hosts = ENV['fab_hosts']
+    user_envs = {}
+    host_connections_by_user = {}
+    
+    # grok fab_hosts into who connects to where
+    for host in fab_hosts:
+        if '@' in host:
+            user, _, host_and_port = partition(host, '@')
+        else:
+            user, host_and_port = None, host
+        hostname, _, port = partition(host_and_port, ':')
+        user = user or username
+        port = int(port or def_port)
+        if user is not '' and user not in user_envs:
+            user_envs[user] = {'fab_user': user}
+        conn = HostConnection(hostname, port, ENV, user_envs[user])
+        if user not in host_connections_by_user:
+            host_connections_by_user[user] = [conn]
+        else:
+            host_connections_by_user[user].append(conn)
+    
+    # Print and establish connections
+    for user, host_connections in host_connections_by_user.iteritems():
+        user_env = dict(ENV)
+        user_env.update(user_envs[user])
+        print(_lazy_format("Logging into the following hosts as $(fab_user):",
+            user_env))
+        print(_indent('\n'.join(map(str, host_connections))))
+        map(HostConnection.connect, host_connections)
+        CONNECTIONS += host_connections
     set(fab_connected=True)
 
 def _disconnect():
     "Disconnect all clients."
     global CONNECTIONS
-    for host, client in CONNECTIONS:
-        client.close()
+    map(HostConnection.disconnect, CONNECTIONS)
     CONNECTIONS = []
 
 def _lazy_format(string, env=ENV):
@@ -947,7 +1074,7 @@ def _lazy_format(string, env=ENV):
 def _on_hosts_do(fn, *args, **kwargs):
     """
     Invoke the given function with hostname and client parameters in
-    accord with the current fac_mode strategy.
+    accord with the current fab_mode strategy.
     
     fn should be a callable taking these parameters:
         hostname : str
@@ -959,33 +1086,53 @@ def _on_hosts_do(fn, *args, **kwargs):
     strategy = ENV['fab_mode']
     if strategy in STRATEGIES:
         strategy_fn = STRATEGIES[strategy]
-        successful = strategy_fn(fn, *args, **kwargs)
-        if not successful:
-            print("Operation failed: " + fn.__name__)
-            sys.exit(1)
+        strategy_fn(fn, *args, **kwargs)
     else:
         print("Unsupported fab_mode: %s" % strategy)
         print("Supported modes are: %s" % (', '.join(STRATEGIES.keys())))
         sys.exit(1)
 
+def _try_run_operation(fn, host, client, env, *args, **kwargs):
+    """
+    Used by strategies to attempt the execution of an operation, and handle
+    any failures appropreately.
+    """
+    err_msg = "The $(fab_current_operation) operation failed on $(fab_host)"
+    success = False
+    try:
+        success = fn(host, client, env, *args, **kwargs)
+    except SystemExit:
+        raise
+    except BaseException, e:
+        _fail(kwargs, err_msg + ':\n' + _indent(str(e)), env)
+    if not success:
+        _fail(kwargs, err_msg + '.', env)
+
 def _confirm_proceed(exec_type, host, kwargs):
     if 'confirm' in kwargs:
         infotuple = (exec_type, host, _lazy_format(kwargs['confirm']))
         question = "Confirm %s for host %s: %s [yN] " % infotuple
         answer = raw_input(question)
-        return answer in 'yY'
+        return answer and answer in 'yY'
     return True
 
-def _get_failcode(kvarg_map, default='abort'):
+def _fail(kwargs, msg, env=ENV):
+    # Get failure code
     codes = {
-        'ignore': 1,
-        'warn': 2,
-        'abort': 3,
+        'ignore': (1, ''),
+        'warn': (2, 'Warning: '),
+        'abort': (3, 'Error: '),
     }
-    if 'fail' in kvarg_map:
-        return codes[kvarg_map['fail']]
-    else:
-        return codes[default]
+    code, msg_prefix = codes[env['fab_fail']]
+    if 'fail' in kwargs:
+        code, msg_prefix = codes[kwargs['fail']]
+    # If warn or above, print message
+    if code > 1:
+        print(msg_prefix + _lazy_format(msg, env))
+        # If abort, also exit
+        if code > 2:
+            sys.exit(1)
+
 
 def _start_outputter(prefix, channel):
     def outputter():
@@ -1053,8 +1200,8 @@ def main():
     try:
         try:
             print(__greeter__ % ENV)
-            fabfile = _pick_fabfile()
             _load_default_settings()
+            fabfile = _pick_fabfile()
             load(fabfile, fail='warn')
             commands = _parse_args(args)
             _validate_commands(commands)
